/*********************************************************************
 * BOPO - BOoost POwer Supply Program
 *********************************************************************
 * FileName:        main.c
 * Dependencies:    See INCLUDES section below
 * Processor:       18F2455
 * Compiler:        C18 2.30.01+
 * Author:          Robert B. Lang 
 * This software works with the hardware described in the article
 * Implementing a Boost Power Supply on a PIC Microprocessor
 * by Robert Lang.  
 * Caution: This circuit/program can generate 5-210 volts DC. 
 * 20Mhz crystal driving 18F2455 PIC at 48 Mhz. 
 *
 * THIS SOFTWARE IS PROVIDED IN AN AS IS CONDITION. NO WARRANTIES,
 * WHETHER EXPRESS, IMPLIED OR STATUTORY, INCLUDING, BUT NOT LIMITED
 * TO, IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE APPLY TO THIS SOFTWARE. THE AUTHOR SHALL NOT,
 * IN ANY CIRCUMSTANCES, BE LIABLE FOR SPECIAL, INCIDENTAL OR
 * CONSEQUENTIAL DAMAGES, FOR ANY REASON WHATSOEVER.
 *
/** I N C L U D E S **********************************************************/
#include <timers.h>
#include <pwm.h>
#include <adc.h>
#include <delays.h>
#include <stdio.h>
#include "lcd.h"
//
#pragma config PLLDIV=5, CPUDIV=OSC1_PLL2, USBDIV=2
#pragma config IESO=OFF, FCMEM=OFF, FOSC=HSPLL_HS 
#pragma config PWRT=OFF, BOR=ON, BORV=21, VREGEN=ON, MCLRE=ON
#pragma config PBADEN=OFF, STVREN=ON, LVP=OFF, ICPRT=OFF 
#pragma config XINST=OFF, DEBUG=OFF, WDT=OFF

// FUNCTION PROTOTYPES
void ADCGet(void);
void ONTIME (void);
void timer_handler(void);
/** V A R I A B L E S ********************************************************/
#pragma udata
char usb_led_status;
static volatile unsigned long VDES=0;
static volatile unsigned long VOUT=0;
unsigned long PERIOD=32000;  // period can be between 1 and 338640 nanoseconds
// this value is 32 microseconds.  Too large of period will cause overheating.                           
unsigned long MAXVOLTS=211;  // maximum of 211 volts
// change following line if execution time or Timer2 settings changes
unsigned long PERIOD_TIME_UNIT=1332; // time in nsec corresponding to one unit in period (PR2 register)
// 20.8 nanoseconds* Timer2 prescaler (16)* 4 =~1332
unsigned long On_Time;
unsigned long UPPERLIMIT;  // LIMIT FOR On_TIME
unsigned long LOWERLIMIT;  // LIMIT FOR On_TIME
//float PREV_INTEGRAL;
//float PREV_ERROR;
static char buffer[20];  // WORKING BUFFER

//#pragma udata usbram7a=0x700  //SET UP MEMORY BLOCK FOR SERIAL BUFFERS
//volatile char serial_input_buffer[128];
//volatile char serial_output_buffer[128];

#pragma code int_vector=0x08   // SET INTERRUPT VECTOR LOCATION
void int_vector (void)
{
  _asm goto timer_handler _endasm
}
#pragma code
#pragma interrupt timer_handler
//**************************************************************************
// Timer interrupts are handled by this routine 
//**************************************************************************
void timer_handler(void)
{ unsigned char On_Timechar;
 ADCGet(); // Measure VDES and VOUT
 ONTIME(); // Calculate on time for desired voltage (interation) and store in On_Time
 On_Timechar=On_Time;  // convert to unsigned char
 SetDCPWM1(On_Timechar); // Set desired PWM on time
  INTCONbits.TMR0IF=0;  // reset timer0 interrupt flag
 WriteTimer0(161);	 // reload value for desired delay
}

#pragma code
void EnableInterrupts(void)
//****************************************************************************
//   EnableInterrupts(void) enables Timer interrupts. 
//****************************************************************************
{ unsigned char PERIODchar;
	unsigned long TEMP;
// HPPLL clock runs at 48 MHz=FOSC
// PERIOD=TOSC=1/48Mhz=20.8nsec, Timer clock at 48/4=12Mhz
// Timer0 increments every .0833 microseconds*128 =10.66 microseconds
// setting up for 1000 microseconds (test)
    OpenTimer0(TIMER_INT_ON & T0_8BIT & T0_SOURCE_INT & T0_PS_1_128);
    WriteTimer0(161);
    INTCONbits.GIE=1; // enable interrupts
// Timer2 is used by the PWM function and must be opened    	
    OpenTimer2(TIMER_INT_OFF &  T2_PS_1_16);
    
// Set the period for the PWM
//   PWM_PERIOD=[MYPERIOD+1]*4*TOSC*TMR2_PRESCALER
//   PWM_PERIOD=[MYPERIOD+1]*4*20.8nsec*16=[MYPERIOD+1]*1.331e-6 seconds
//   From above each Unit of MYPERIOD increases PWM_PERIOD by 1.331 micro seconds
//OpenPWM1(MYPERIOD);
TEMP=PERIOD/ PERIOD_TIME_UNIT;
PERIODchar=TEMP-1;//PR2 register contains one less than PERIOD calc 
// The next two lines set limits for ONTIME
LOWERLIMIT=TEMP/4; // calculate 1/16
UPPERLIMIT=LOWERLIMIT*15;  // calculate 15/16
OpenPWM1(PERIODchar); // set period to nanoseconds specified in PERIOD
}

void ONTIME (void)
//****************************************************************************
//   This routine calculates the on time based on VDES and VOUT. 
//****************************************************************************
{  
//	 float ERROR;
     //	 float Pk;
//	 float Ik;
//	 float Dk;
//	ERROR=VDES-VOUT;
//	Pk=Kp*ERROR;
//	Ik=PREV_INTEGRAL + Ki * ERROR;
//	Dk=Kd*(PREV_ERROR -ERROR);
//	// save data needed for next iteration
//	PREV_INTEGRAL=Ik;
//	PREV_ERROR=ERROR;
//	On_Time=On_Time*(Pk+Ik+Dk);
// A SIMPLIFIED CORRECTION ALGORITHM
    signed long ERROR;
	ERROR=VDES-VOUT;
	// We will decrease the PWM  on time when the the desired voltage is greater than the 
	// because decreasing the on time causes the coil to be grounded longer causing the
	// output voltage to be higher.   
    if ( ERROR > 0) On_Time=On_Time+1;
    if (ERROR < 0)  On_Time=On_Time-1;
    if (On_Time==-1) On_Time=0;
    // The on time must be a 10 bit unsigned number
    if (On_Time >1023)   On_Time=1023 ;
// CHECK VALID ONTIME RANGE
     if (On_Time> UPPERLIMIT) On_Time=UPPERLIMIT; // keep upper limit to prevent coil overheat 
     // and entry into continuous current mode.
}
	
void ADCInit(void)
//****************************************************************************
//   Initialize Analog to Digital Conversion
//****************************************************************************
{ unsigned char config;
  unsigned char config2;
  unsigned char portconfig;
  #define TRIS_AN0  DDRAbits.RA0    /* TRIS for CHANNEL 0 */ 
  #define TRIS_AN1  DDRAbits.RA1    /* TRIS for CHANNEL 1 */ 
  #define TRIS_AN2  DDRAbits.RA2    /* TRIS for CHANNEL 2 */
  // MAKE SURE AN0 AND AN1 ARE INPUTS FOR A/D TO WORK
   TRIS_AN0=1;
   TRIS_AN1=1;
   TRIS_AN2=1; 	
// config=select ADC clock source, justification and conversion time
// config2=select channel and interrupt
// portconfig= contents of ADCON1 register=12 enables AN0, AN1 AND AN2 as analog.	
// AN2 is not currently used.
   config=ADC_FOSC_64  & ADC_RIGHT_JUST & ADC_20_TAD;
   config2=ADC_CH0 & ADC_VREFPLUS_VDD & ADC_VREFMINUS_VSS & ADC_INT_OFF;
   portconfig=12;  
   OpenADC(config, config2, portconfig);
 }

void ADCGet(void)
//****************************************************************************
//   Perform Analog to Digital Conversion for VDES and VOUT
//****************************************************************************
{ 
  SetChanADC(ADC_CH0);  // Select channel 0 (VOUT)
  ConvertADC();      // Start A/D conversion
  while(BusyADC());  // Wait for conversion to complete
  VOUT=ReadADC();    // get converted value

    // 0-5 volts=0 to 1023 counts
  SetChanADC(ADC_CH1);  // Select channel 1 (VDES)
  ConvertADC();      // Start A/D conversion
  while(BusyADC());  // Wait for conversion to complete
  VDES=ReadADC();    // get converted value
// 0-5 volts=0 to 1023 counts
}

// SET UP DELAYS USED BY LCD LIBRARY FUNCTIONS
void  DelayFor18TCY(void)
  {
 Delay10TCYx(2); // delay 20 instruction cycles 20*83nsec=1660nsec
  } 
void  DelayXLCD(void)
  {
  Delay10KTCYx(60); // delay 60,000 instruction cycles (5 msec)
  }
void  DelayPORXLCD(void)
  {
  Delay10KTCYx(180); // delay 180,000 instruction cycles (15 msec)
  }  

void SendLCDCommand(unsigned char MYCOMMAND)
// The sendlcdcommand function sends a command to the lcd
{ 
  TRIS_DATA_PORT = 0;           // Make port output
  RS_PIN=0;  // Command mode
  DATA_PORT=MYCOMMAND;
  RW_PIN = 0; // Set control bit for write
  DelayFor18TCY(); // allow control signals and data to settle
  E_PIN = 1;                      // Clock data into LCD
 DelayFor18TCY();
 E_PIN = 0;
 RS_PIN = 0;                     // Reset control bits
}

void SendLCDData(char MYDATA)
// This routine sends a data character to the LCD
{
        TRIS_DATA_PORT = 0;             // Make port output
        DATA_PORT = MYDATA;               // Write data to port
        RS_PIN = 1;   // Data mode
        RW_PIN = 0;   // write mode
        DelayFor18TCY(); // allow control signals and data to settle
        E_PIN = 1;                      // Clock data into LCD
        DelayFor18TCY();
        E_PIN = 0;
        RS_PIN = 0;                     // Reset control bits
}

void WaitWhileLCDBusy(void)
// The WaitWhileLCDBusy function returns if not busy
{       unsigned char BUSY;
LOOP:   DelayXLCD();;   // WAIT 5MSEC BETWEEN EACH BUSY CHECK
        TRIS_DATA_PORT=0xFF; // set port for input
        RW_PIN = 1;         // Set the control bits for read
        RS_PIN = 0;
        DelayFor18TCY();  // Allow time for control pins to settle
        E_PIN = 1;      // Enable transfer
        DelayFor18TCY();  // Allow time for data to settle
        BUSY=DATA_PORT;
        E_PIN = 0;              // Reset enable
        if(BUSY&0x80){goto LOOP;}      // BUSYBIT SET?
	    RW_PIN = 0;   // RESET CONTROL LINES                   
        RS_PIN = 0;    
 }

void LCDErase(void)
// The erase function erases the LCD
{
WaitWhileLCDBusy(); //wait till lcd not busy
SendLCDCommand(0x01);  // erase command
WaitWhileLCDBusy(); //wait till lcd not busy
}

void LCD_Display_Rom(const rom char *buffer)
// This routine displays string in ROM on LCD
{       
	 	    WaitWhileLCDBusy();  // Make sure device is not busy before starting
        while(*buffer)                  // Write data to LCD up to null
        {   WaitWhileLCDBusy();      // Wait while LCD is busy
                SendLCDData(*buffer); // Write character to LCD
                buffer++;               // Increment buffer
        }
        return;
}

void LCD_Display_Ram(char *buffer)
// This routine displays string in RAM on LCD
{
	   WaitWhileLCDBusy();  // Make sure device is not busy before starting
        while(*buffer)                  // Write data to LCD up to null
        {       WaitWhileLCDBusy();      // Wait while LCD is busy
                SendLCDData(*buffer); // Write character to LCD
                buffer++;               // Increment buffer
        }
}
  
void LCDInit(void)
//****************************************************************************
//   Initialize LCD
//   LCD library functions are for the control of an
//   LCD controller.  The PIC pin definitions used below 
//   are defined in lcd.h.
//   DATA LINES = RB0-7
//   RS PIN= PIC RA2 to LCD connector Pin4
//   RW PIN= PIC RA3 to LCD connector Pin5
//   E PIN= PIC RA4 to LCD connector Pin6
//****************************************************************************
{
        TRIS_DATA_PORT=0;
        TRIS_RW = 0;                    // All control signals made outputs
        TRIS_RS = 0;
        TRIS_E = 0;
        RW_PIN = 0;                     // R/W pin made low
        RS_PIN = 0;                     // Register select pin made low
        E_PIN = 0;                      // Clock pin made low
        // Delay for 15ms to allow for LCD Power on reset
    	 DelayPORXLCD();
// send hex 30 three times to initialize LCD
        SendLCDCommand(0x30);
        DelayXLCD();
        SendLCDCommand(0x30);
        DelayXLCD();
        SendLCDCommand(0x30);
        DelayXLCD();
// busy flag is now operational and will be used to insure that LCD actions are complete
        SendLCDCommand(0x08);  // display off 
           WaitWhileLCDBusy();
        SendLCDCommand(0x06);  // auto increment and no shifting
             WaitWhileLCDBusy();
        SendLCDCommand(0x0E);  // display and cursor on and blink off
             WaitWhileLCDBusy();                
        LCDErase();
     	LCD_Display_Rom( "BOOST POWER SUPPLY" );
     	 Delay10KTCYx(0); // delay ~2.6 million cycles
         Delay10KTCYx(0); // delay ~2.6 million cycles
 }

//****************************************************************************
//   Displays Voltages on LCD 
//****************************************************************************
void DisplayVoltages(void)
{
    unsigned int DES_INT;
    unsigned int DES_TENTHS;
    unsigned int OUT_INT;
    unsigned int OUT_TENTHS;
    long IVOUT;
    long IVDES;
    long DVDES;
    long DVOUT;
// convert 0-1023 to 0.0 to MAXVOLTS.0 
    IVDES=(VDES*MAXVOLTS)/1023; // get integer component of voltage
    DES_INT=IVDES; // convert from long to int
    DVDES=(VDES*MAXVOLTS*10)/1023-(IVDES *10); // isolate 10ths digit
    DES_TENTHS=DVDES; // convert from long to int
// convert 0-1023 to 0.0 to MAXVOLTS.0     
    IVOUT=(VOUT*MAXVOLTS)/1023; // get integer component of voltage
    OUT_INT=IVOUT; // convert from long to int
    DVOUT=(VOUT*MAXVOLTS*10)/1023-(IVOUT *10); // isolate 10ths digit
    OUT_TENTHS=DVOUT; // convert from long to int
   sprintf(buffer,"Vd=%3d.%d Vout=%3d.%1d",DES_INT,DES_TENTHS,OUT_INT,OUT_TENTHS); // FORMAT STRING FOR LCD with 10ths
  // negative numbers indicate faulty conversion, try again
  if (DES_INT < 0) return;
  if (OUT_INT < 0) return;
  if (DES_TENTHS < 0) return;
  if (OUT_TENTHS < 0) return;
   LCD_Display_Ram( buffer );  //UPDATE LCD
}

//****************************************************************************
//   Displays ONTIME on LCD 
//****************************************************************************
void DisplayONTIME(void)
{
    unsigned long TEMP;
    unsigned int TIME;
    TEMP=On_Time; // convert to long
    TEMP=TEMP*PERIOD_TIME_UNIT;
    TEMP=TEMP/4; // convert to nanoseconds
    TIME=TEMP; // convert to int
    sprintf(buffer,"On Time=%5d nsec",TIME); // FORMAT STRING FOR LCD
    LCD_Display_Ram( buffer );  //UPDATE LCD
}

//****************************************************************************
//   Displays PERIOD on LCD 
//****************************************************************************
void DisplayPeriod(void)
{
    unsigned int TIME;
    TIME=PERIOD; // convert to int  
    sprintf(buffer,"Period=%5d nsec",TIME); // FORMAT STRING FOR LCD
    LCD_Display_Ram( buffer );  //UPDATE LCD
}

void main(void)
//****************************************************************************
// main(void) the main program
//****************************************************************************
{ 
//	PREV_INTEGRAL=0.;
//	PREV_ERROR=0.;
	On_Time=0; // initialize ONTIME minimum
	
	ADCInit(); // set an0 and an 1 to analog, all other porta are digital
	EnableInterrupts();
	DDRCbits.RC0=0; // MAKE RC0 AN OUTPUT BY SETTING DATA DIRECTION REGISTER BIT TO 0
	PORTCbits.RC0=1; // TURN ON LED
    LCDInit();

// and loop forever...    
     while(1) 
    {
  Delay10KTCYx(0); // delay ~2.6 million cycles
  Delay10KTCYx(0); // delay ~2.6 million cycles
  LCDErase();// erase the LCD
  PORTCbits.RC0=0; // TURN OFF LED while VDES and VOUT are displayed	  
  DisplayVoltages(); // display voltages VDES and VOUT
 Delay10KTCYx(0); // delay ~2.6 million cycles
 Delay10KTCYx(0); // delay ~2.6 million cycles
 Delay10KTCYx(0); // delay ~2.6 million cycles
 Delay10KTCYx(0); // delay ~2.6 million cycles
  LCDErase();// erase the LCD
 PORTCbits.RC0=1; // TURN ON LED while ontime is displayed   
 DisplayONTIME(); // display ONTIME
 Delay10KTCYx(0); // delay ~2.6 million cycles
 Delay10KTCYx(0); // delay ~2.6 million cycles
 Delay10KTCYx(0); // delay ~2.6 million cycles
 Delay10KTCYx(0); // delay ~2.6 million cycles
 LCDErase();// erase the LCD  
 DisplayPeriod(); // display period
 	}     
} //end main

